# fields
from abc import ABC
from datetime import date
from enum import Enum
from typing import Dict, List, Optional, Union

from slack_ui.blocks import Block


class Field(Block, ABC):
    name: str
    label: str

    def __init__(self, label: str):
        self.label = label

    @property
    def action_id(self) -> str:
        return "%(name)s-action" % {"name": self.name}


class TextField(Field):
    def slack_block(self) -> dict:
        return {
            "type": "input",
            "element": {
                "type": "plain_text_input",
                "action_id": self.action_id,
            },
            "label": {"type": "plain_text", "text": self.label, "emoji": True},
        }


class SelectFormEnum(Enum):
    @classmethod
    def get_select_options(cls):
        return {a.key: a.label for a in cls}

    @property
    def key(self):
        try:
            return self.value[0] if self.is_enumerated else self.value
        except (KeyError, AttributeError, IndexError):
            return self.name

    @property
    def label(self):
        try:
            return self.value[1] if self.is_enumerated else self.value
        except (KeyError, AttributeError, IndexError):
            return self.name

    @property
    def is_enumerated(self) -> bool:
        return isinstance(self.value, (tuple, list))


class StaticSelectField(Field):
    options: Union[Dict[str, str], SelectFormEnum] = {}
    placeholder: str = "Select an item"

    def __init__(self, label: str, options=None, placeholder: str = "Select an item"):
        super().__init__(label)
        self.options = options if options else {}
        self.placeholder = placeholder

    def slack_block(self) -> dict:
        return {
            "type": "input",
            "element": {
                "type": "static_select",
                "placeholder": {
                    "type": "plain_text",
                    "text": self.placeholder,
                    "emoji": True,
                },
                "options": [
                    {
                        "text": {
                            "type": "plain_text",
                            "text": self.get_options()[key],
                            "emoji": True,
                        },
                        "value": key,
                    }
                    for key in self.get_options()
                ],
                "action_id": self.action_id,
            },
            "label": {"type": "plain_text", "text": self.label, "emoji": True},
        }

    def get_options(self) -> dict:
        # noinspection PyTypeChecker
        return (
            self.options.get_select_options()
            if issubclass(self.options, SelectFormEnum)
            else self.options
        )


class DateField(Field):
    initial_date: Optional[date] = None
    placeholder: str = "Select a date"

    def __init__(
        self,
        label: str,
        initial_date: Optional[date] = None,
        placeholder: str = "Select a date",
    ):
        super().__init__(label)
        self.initial_date = initial_date if initial_date else None
        self.placeholder = placeholder

    def slack_block(self) -> dict:
        return {
            "type": "input",
            "element": {
                "type": "datepicker",
                "placeholder": {
                    "type": "plain_text",
                    "text": self.placeholder,
                    "emoji": True,
                },
                "action_id": self.action_id,
                **(
                    {"initial_date": self.initial_date.strftime("%Y-%m-%d")}
                    if self.initial_date
                    else {}
                ),
            },
            "label": {"type": "plain_text", "text": self.label, "emoji": True},
        }


class FormMetadata(type):
    def __new__(mcs, name, bases, attrs):
        # Collect fields from current class and remove them from attrs.
        attrs["declared_fields"] = {
            key: attrs.pop(key)
            for key, value in list(attrs.items())
            if isinstance(value, Field)
        }

        new_class = super().__new__(mcs, name, bases, attrs)

        # Walk through the MRO.
        declared_fields = {}
        for base in reversed(new_class.__mro__):
            # Collect fields from base class.
            if hasattr(base, "declared_fields"):
                declared_fields.update(base.declared_fields)

            # Field shadowing.
            for attr, value in base.__dict__.items():
                if value is None and attr in declared_fields:
                    declared_fields.pop(attr)

        new_class.base_fields = declared_fields
        new_class.declared_fields = declared_fields

        return new_class


class Form(metaclass=FormMetadata):
    def get_fields(self) -> List[Field]:
        fields: List[Field] = []
        for field_name in self.declared_fields:
            field = self.declared_fields[field_name]
            field.name = field_name
            fields.append(field)
        return fields

    def slack_view(self) -> List[Dict]:
        return [field.slack_block() for field in self.get_fields()]



import json
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Literal, Optional


class Block(ABC):
    """Abstract base class representing a Slack block.

    This class defines the interface for all Slack blocks.

    Methods:
        slack_block: Abstract method that returns a dict representing the Slack block.
    """
    @abstractmethod
    def slack_block(self) -> dict:
        pass


class Element(Block, ABC):
    pass


class Button(Element):
    style: Optional[Literal["primary", "danger"]]
    action_id: str
    value: Any
    text: str

    def __init__(
        self,
        action_id: str,
        text: str,
        value: Any,
        style: Optional[Literal["primary", "danger"]] = None,
    ):
        assert style in [None, "primary", "danger"]

        self.action_id = action_id
        self.value = value
        self.style = style
        self.text = text

    def slack_block(self) -> dict:
        return {
            "type": "button",
            "text": {"type": "plain_text", "text": self.text},
            "value": json.dumps(self.value),
            "action_id": self.action_id,
            **({"style": self.style} if self.style else {}),
        }


class ActionsBlock(Block):
    elements: List[Element]

    def __init__(self, elements: Optional[List[Element]] = None):
        self.elements = elements if elements else None

    def slack_block(self) -> dict:
        return {
            "type": "actions",
            "elements": [element.slack_block() for element in self.elements],
        }


class HeaderBlock(Block):
    text: str

    def __init__(self, text: str) -> None:
        assert (
            len(text) <= 150
        ), "[Header Block] Text can't be longer then 150 characters"

        self.text = text

    def slack_block(self) -> dict:
        return {
            "type": "header",
            "text": {
                "type": "plain_text",
                "text": self.text,
            },
        }


class TextBlock(Block):
    text: str

    def __init__(self, text: str) -> None:
        assert (
            len(text) <= 3000
        ), "[Text Block] Text can't be longer then 3000 characters"

        self.text = text

    def slack_block(self) -> dict:
        return {
            "type": "section",
            "text": self.slack_field(),
        }

    def slack_field(self) -> Dict[str, str]:
        return {"type": "mrkdwn", "text": self.text}

    def __str__(self):
        return f"TextBlock({self.text})"


class SectionBlock(Block):
    fields: List[str] = []
    text: Optional[str]

    def __init__(self, text: Optional[str], fields: Optional[List[str]] = None) -> None:
        assert (
            text is None or len(text) <= 3000
        ), "[Section Block] Text can't be longer then 3000 characters"
        assert (
            max(len(s) for s in fields) <= 2000
        ), "[Section Block] No one text in fields can't be longer then 2000 characters"
        assert len(fields) <= 10, "[Section Block] Max number of fields is 10"

        self.fields = fields if fields is not None else []
        self.text = text

    def slack_block(self) -> dict:
        return {
            "type": "section",
            **({"text": {"type": "mrkdwn", "text": self.text}} if self.text else {}),
            **(
                {"fields": [{"type": "mrkdwn", "text": field} for field in self.fields]}
                if len(self.fields)
                else {}
            ),
        }


class ContextBlock(Block):
    context: List[TextBlock]

    def __init__(self, context: List[TextBlock]) -> None:
        assert len(context) <= 10, "[Context Block] Max number of fields is 10"

        self.context = context

    def slack_block(self) -> dict:
        return {
            "type": "context",
            "elements": [field.slack_field() for field in self.context],
        }


class DividerBlock(Block):
    def slack_block(self) -> dict:
        return {"type": "divider"}



from __future__ import annotations

from slack_ui.blocks import Block
from slack_ui.domain import BlocksList, DeepBlocksList


def flat_blocks_list(*blocks_list: Block | DeepBlocksList) -> BlocksList:
    """
    Flatten a list of blocks and deep blocks lists into a single list.

    :param blocks_list: A variable number of Block or DeepBlocksList objects.
    :type blocks_list: tuple[Block | DeepBlocksList]

    :return: A list containing all the blocks in the input, flattened.
    :rtype: list[Block]
    """
    result: BlocksList = []

    for block in blocks_list:
        if isinstance(block, Block):
            result.append(block)
        elif isinstance(block, (list, tuple)):
            result.extend(flat_blocks_list(*block))

    return result



from __future__ import annotations

from typing import List, Union

from slack_ui.blocks import Block

BlocksList = List[Block]
DeeperBlocksList = List[Union[Block, BlocksList]]
DeepBlocksList = List[Union[Block, DeeperBlocksList]]



from typing import List, Optional

from slack_ui.blocks import Block
from slack_ui.domain import BlocksList
from slack_ui.forms import Form


class Message:
    blocks: BlocksList = []

    def slack_view(self) -> List[dict]:
        return self.generate_blocks()

    def add_block(self, block: Block):
        self.blocks.append(block)

    def add_blocks(self, blocks: BlocksList):
        for block in blocks:
            self.add_block(block)

    def generate_blocks(self) -> List[dict]:
        return [b.slack_block() for b in self.blocks]


class Modal(Message):
    title: str
    callback_id: str

    def __init__(self, title: str, callback_id: str) -> None:
        self.callback_id = callback_id
        self.title = title

    def slack_view(self) -> dict:
        return {
            "type": "modal",
            "callback_id": self.callback_id,
            "title": {"type": "plain_text", "text": self.title},
            "blocks": self.generate_blocks(),
            **self.additional_blocks(),
        }

    def additional_blocks(self) -> dict:
        return {}


class FormModal(Modal):
    form: Form
    submit_text: str

    def __init__(
        self,
        title: str,
        callback_id: str,
        form: Form,
        submit_text: Optional[str] = "Submit",
    ) -> None:
        super().__init__(title, callback_id)
        self.submit_text = submit_text
        self.form = form

    def generate_blocks(self) -> List[dict]:
        return self.form.slack_view()

    def additional_blocks(self) -> dict:
        return {"submit": {"type": "plain_text", "text": self.submit_text}}
